@@ -3,16 +3,16 @@
  *	This file contains Magic command routines for those commands
  *	that are valid in all windows.
  *
- *     ********************************************************************* 
- *     * Copyright (C) 1985, 1990 Regents of the University of California. * 
- *     * Permission to use, copy, modify, and distribute this              * 
- *     * software and its documentation for any purpose and without        * 
- *     * fee is hereby granted, provided that the above copyright          * 
- *     * notice appear in all copies.  The University of California        * 
- *     * makes no representations about the suitability of this            * 
- *     * software for any purpose.  It is provided "as is" without         * 
- *     * express or implied warranty.  Export of this software outside     * 
- *     * of the United States of America may require an export license.    * 
+ *     *********************************************************************
+ *     * Copyright (C) 1985, 1990 Regents of the University of California. *
+ *     * Permission to use, copy, modify, and distribute this              *
+ *     * software and its documentation for any purpose and without        *
+ *     * fee is hereby granted, provided that the above copyright          *
+ *     * notice appear in all copies.  The University of California        *
+ *     * makes no representations about the suitability of this            *
+ *     * software for any purpose.  It is provided "as is" without         *
+ *     * express or implied warranty.  Export of this software outside     *
+ *     * of the United States of America may require an export license.    *
  *     *********************************************************************
  */
 
@@ -190,7 +190,7 @@ windCaptionCmd(w, cmd)
 	TxError("Usage: %s [on|off]\n", cmd->tx_argv[0]);
 	return;
 }
-
+
 /*
  * ----------------------------------------------------------------------------
  *
@@ -209,7 +209,7 @@ windCaptionCmd(w, cmd)
  *
  * Side effects:
  *	The view in the window underneath the cursor is changed
- *	to center the point underneath the cursor.  
+ *	to center the point underneath the cursor.
  *
  * ----------------------------------------------------------------------------
  */
@@ -298,7 +298,7 @@ windCenterCmd(w, cmd)
     WindMove(w, &newArea);
 }
 
-
+
 /*
  * ----------------------------------------------------------------------------
  * windCloseCmd --
@@ -363,7 +363,7 @@ windCloseCmd(w, cmd)
  *	None.
  *
  * Side effects:
- *	
+ *
  *
  * ----------------------------------------------------------------------------
  */
@@ -421,7 +421,7 @@ windCrashCmd(w, cmd)
     TxFlush();
     niceabort();
 }
-
+
 
 /*
  * ----------------------------------------------------------------------------
@@ -507,7 +507,7 @@ windCursorCmd(w, cmd)
 	    ToolSnapToGrid(w, &p_out, (Rect *)NULL);
     }
 
-    /* Transform the result to declared units with option "lambda" or "grid" */ 
+    /* Transform the result to declared units with option "lambda" or "grid" */
     switch (resulttype) {
 	case -2:
 	case -1:
@@ -538,7 +538,7 @@ windCursorCmd(w, cmd)
 
 #ifdef MAGIC_WRAPPER
     listxy = Tcl_NewListObj(0, NULL);
-    if ((cursx == round(cursx)) && (cursy == round(cursy))) 
+    if ((cursx == round(cursx)) && (cursy == round(cursy)))
     {
 	Tcl_ListObjAppendElement(magicinterp, listxy, Tcl_NewIntObj((int)cursx));
 	Tcl_ListObjAppendElement(magicinterp, listxy, Tcl_NewIntObj((int)cursy));
@@ -575,14 +575,14 @@ windDebugCmd(w, cmd)
 {
     if (cmd->tx_argc != 1) goto usage;
     windPrintCommands = !windPrintCommands;
-    TxError("Window command debugging set to %s\n", 
+    TxError("Window command debugging set to %s\n",
 	(windPrintCommands ? "TRUE" : "FALSE"));
     return;
 
 usage:
     TxError("Usage:  *winddebug\n");
 }
-
+
 /*
  * ----------------------------------------------------------------------------
  * windDumpCmd --
@@ -606,7 +606,7 @@ windDumpCmd(w, cmd)
 }
 
 #ifndef MAGIC_WRAPPER
-
+
 /*
  * ----------------------------------------------------------------------------
  *
@@ -640,7 +640,7 @@ windEchoCmd(w, cmd)
 	else
 	    TxPrintf("%s", cmd->tx_argv[i]);
     }
-    
+
     if (newline)
 	TxPrintf("\n");
     TxFlush();
@@ -648,7 +648,7 @@ windEchoCmd(w, cmd)
 
 #endif
 
-
+
 /*
  * ----------------------------------------------------------------------------
  *
@@ -695,7 +695,7 @@ windFilesCmd(w, cmd)
 		case S_IFSOCK: {type = "socket"; break;}
 		default: {type = "unknown"; break;}
 	    }
-	    TxError("file descriptor %d: open  (type: '%s', inode number %ld)\n", 
+	    TxError("file descriptor %d: open  (type: '%s', inode number %ld)\n",
 		fd, type, buf.st_ino);
 	    open++;
 	}
@@ -703,7 +703,7 @@ windFilesCmd(w, cmd)
     TxError("%d open files, %d unopened file descriptors left\n", open, unopen);
 }
 
-
+
 /*
  * ----------------------------------------------------------------------------
  *
@@ -734,7 +734,7 @@ windGrowCmd(w, cmd)
     WindFullScreen(w);
 }
 
-
+
 /*
  * ----------------------------------------------------------------------------
  * windGrstatsCmd --
@@ -767,7 +767,7 @@ windGrstatsCmd(w, cmd)
 	return;
     }
 
-    if (!StrIsInt(cmd->tx_argv[1]) || 
+    if (!StrIsInt(cmd->tx_argv[1]) ||
 	(cmd->tx_argc == 3 && !StrIsInt(cmd->tx_argv[2])))
     {
 	TxError("Count & style must be numeric\n");
@@ -832,14 +832,14 @@ windGrstatsCmd(w, cmd)
     us = tdelta.tms_utime * (1000000 / 60);
     usPerRect = us / MAX(1, GrNumClipBoxes);
     rectsPerSec = 1000000 / MAX(1, usPerRect);
-    TxPrintf("[%s]\n%d rectangles, %d uS, %d uS/rectangle, %d rects/sec\n", 
+    TxPrintf("[%s]\n%d rectangles, %d uS, %d uS/rectangle, %d rects/sec\n",
 	rstatp, GrNumClipBoxes, us, usPerRect, rectsPerSec);
 
     if (style >= 0)
 	GrUnlock(w);
 }
 
-
+
 /*
  * ----------------------------------------------------------------------------
  * windHelpCmd --
@@ -863,18 +863,23 @@ windHelpCmd(w, cmd)
     ASSERT(FALSE, windHelpCmd);
 }
 
-static char *logKeywords[] =
-    {
-	"update",
-	0
-    };
-
 /*
  * ----------------------------------------------------------------------------
  * windLogCommandsCmd --
  *
  *	Log the commands and button pushes in a file.
  *
+ * Syntax:
+ *	logcommands start <filename>	Open and start a new command log file
+ *	logcommands stop		End and close a command log file
+ *	logcommands update		Refresh display after each log command
+ *	logcommands suspend		Suspend logging of commands
+ *	logcommands resume		Resume logging of commands
+ *
+ * Legacy syntax:
+ *	logcommands <filename> [update]	Start a new command log file with updating
+ *	logcommands			End and close a command log file
+ *
  * Results:
  *	None.
  *
@@ -883,35 +888,74 @@ static char *logKeywords[] =
  * ----------------------------------------------------------------------------
  */
 
+#define LOG_CMD_START	0	// Create a new log file (default behavior)
+#define LOG_CMD_STOP	1	// End a log file and stop logging.
+#define LOG_CMD_UPDATE	2	// Update display after every logged command.
+#define LOG_CMD_SUSPEND	3	// Suspend command logging.
+#define LOG_CMD_RESUME 	4	// Resume command logging.
+
 void
 windLogCommandsCmd(w, cmd)
     MagWindow *w;
     TxCommand *cmd;
 {
-    char *fileName;
-    bool update;
+    char *fileName = NULL;
+    unsigned char flags = 0;
+    int idx = LOG_CMD_STOP;
+
+    static char *logKeywords[] = {"start", "stop", "update", "suspend", "resume", 0};
 
     if ((cmd->tx_argc < 1) || (cmd->tx_argc > 3)) goto usage;
 
-    update = FALSE;
-
-    if (cmd->tx_argc == 1) 
-	fileName = NULL;
-    else
-	fileName = cmd->tx_argv[1];
-
-    if (cmd->tx_argc == 3) {
-	int i;
-	i = Lookup(cmd->tx_argv[cmd->tx_argc - 1], logKeywords);
-	if (i != 0) goto usage;
-	update = TRUE;
+    if (cmd->tx_argc > 1)
+    {
+	idx = Lookup(cmd->tx_argv[1], logKeywords);
+	if (idx < 0) fileName = cmd->tx_argv[1];
     }
 
-    TxLogCommands(fileName, update);
-    return;
+    if (cmd->tx_argc == 3)
+    {
+	int i;
+
+	if (fileName == NULL)
+	    fileName = cmd->tx_argv[2];
+	else
+	{
+	    /* Legacy behavior:  Allow "logcommands <filename> update"
+	     * to mean "logcommands start <filename> ; logcommands update"
+	     */
+	    i = Lookup(cmd->tx_argv[2], logKeywords);
+	    if (i != LOG_CMD_UPDATE) goto usage;
+	    flags = TX_LOG_UPDATE;
+	    idx = LOG_CMD_START;
+	}
+    }
+
+    switch (idx)
+    {
+	case LOG_CMD_START:
+	    if (fileName == NULL) break;
+	    TxLogStart(fileName, w);
+	    if (flags & TX_LOG_UPDATE)
+	    	TxLogUpdate();
+	    return;
+	case LOG_CMD_STOP:
+	    TxLogStop();
+	    return;
+	case LOG_CMD_UPDATE:
+	    TxLogUpdate();
+	    return;
+	case LOG_CMD_SUSPEND:
+	    TxLogSuspend();
+	    return;
+	case LOG_CMD_RESUME:
+	    TxLogResume();
+	    return;
+    }	
 
 usage:
-    TxError("Usage: %s [filename [update]]\n", cmd->tx_argv[0]);
+    TxError("Usage: %s [start|stop|update|suspend|resume [filename]]\n",
+		cmd->tx_argv[0]);
 }
 
 /*
@@ -985,17 +1029,21 @@ windDoMacro(w, cmd, interactive)
     bool interactive;
 {
     char *cp, *cn;
+    char nulltext[] = "";
     char ch;
-    int ct, argstart;
+    int ct, argstart, verbose;
     bool any, iReturn;
     bool do_list = FALSE;
+    bool do_help = FALSE;
+    bool do_reverse = FALSE;
+    char *searchterm = NULL;
     macrodef *cMacro;
     HashTable *clienttable;
     HashEntry *h;
     HashSearch hs;
     WindClient wc;
 
-    /* If the second argument is a window name, we attempt to	*/
+    /* If the first argument is a window name, we attempt to	*/
     /* retrieve a client ID from it.  This overrides the actual	*/
     /* window the command was called from, so technically we	*/
     /* can define macros for clients from inside other clients.	*/
@@ -1005,40 +1053,71 @@ windDoMacro(w, cmd, interactive)
     /* associated window.					*/
 
     argstart = 1;
-    if (cmd->tx_argc == 1) wc = DBWclientID;  /* Added by NP 11/15/04 */
-
-    if (cmd->tx_argc > 1)
-    {
+    if (cmd->tx_argc == 1)
+	wc = DBWclientID;  /* Added by NP 11/15/04 */
+    else if (cmd->tx_argc > 1)
 	wc = WindGetClient(cmd->tx_argv[1], TRUE);
-	if (wc == (WindClient)NULL)
-	    if (cmd->tx_argc == 4)
-	    {
-		/* This may look odd, but it has a reason.  If	*/
-		/* we don't register a client, such as wind3d 	*/
-		/* when running a non-OpenGL magic, then we	*/
-		/* shuffle off macros for that window type to	*/
-		/* a "fake" NULL client, where it won't be seen	*/
-		/* again, but won't cause an error.		*/
-		wc = 0;
-		argstart = 2;
-	    }
-	    else if (w != NULL)
-		wc = w->w_client;
-	    else
-		wc = DBWclientID;
-	else
-	    argstart = 2;
-    }
 
-    if (cmd->tx_argc > argstart)
+    while (cmd->tx_argc > argstart)
     {
 	if (!strcmp(cmd->tx_argv[argstart], "list"))
 	{
 	    do_list = TRUE;
 	    argstart++;
 	}
+	else if (!strcmp(cmd->tx_argv[argstart], "help"))
+	{
+	    do_help = TRUE;
+	    argstart++;
+	}
+	else if (!strcmp(cmd->tx_argv[argstart], "search"))
+	{
+	    if (cmd->tx_argc > (argstart + 1))
+	    {
+		argstart++;
+		searchterm = cmd->tx_argv[argstart];
+	    }
+	    argstart++;
+	}
+	else if (!strcmp(cmd->tx_argv[argstart], "-reverse"))
+	{
+	    do_reverse = TRUE;
+	    argstart++;
+	}
+	else break;
     }
 
+    /* If client wasn't specified, use window default, else use	*/
+    /* DBW client.						*/
+
+    if (wc == (WindClient)NULL)
+    {
+	if (w != NULL)
+	    wc = w->w_client;
+	else
+	    wc = DBWclientID;
+
+	if (cmd->tx_argc > (argstart + 1))
+	{
+	    /* The first argument, if there is one after resolving	*/
+	    /* all of the optional arugments, should be a key.		*/
+	    /* If it doesn't look like one, then check if the		*/
+	    /* next argument looks like a key, which would indicate	*/
+	    /* an unregistered client as the first argument.  A		*/
+	    /* macro retrieved from an unregistered client returns	*/
+	    /* nothing but does not generate an error.			*/
+
+	    if (MacroKey(cmd->tx_argv[argstart], &verbose) == 0)
+		if (MacroKey(cmd->tx_argv[argstart + 1], &verbose) != 0)
+		{
+		    wc = 0;
+		    argstart++;
+		}
+	}
+    }
+    else
+	argstart++;
+
     if (cmd->tx_argc == argstart)
     {
 	h = HashLookOnly(&MacroClients, wc);
@@ -1064,23 +1143,47 @@ windDoMacro(w, cmd, interactive)
 	    cMacro = (macrodef *) HashGetValue(h);
 	    if (cMacro == (macrodef *)NULL) break;
 	    cn = MacroName((spointertype)h->h_key.h_ptr);
-	    cp = cMacro->macrotext;
+
+	    /* "imacro list" returns only interactive macros. */
+	    if (interactive && !cMacro->interactive) continue;
+
+	    if (do_help)
+		cp = (cMacro->helptext == NULL) ? cMacro->macrotext : cMacro->helptext;
+	    else
+		cp = cMacro->macrotext;
+
+	    if (cp == (char *)NULL) cp = (char *)(&nulltext[0]);
+
+	    if (searchterm != NULL)
+	    {
+		/* Refine results by keyword search */
+		if (!strstr(cp, searchterm)) continue;
+	    }
 
 	    if (do_list)
 	    {
 #ifdef MAGIC_WRAPPER
-		Tcl_AppendElement(magicinterp, cp);
+		// NOTE:  Putting cp before cn makes it easier to
+		// generate a reverse lookup hash table for matching
+		// against menu items, to automatically generate
+		// the "accelerator" text.
+		if (do_reverse)
+		    Tcl_AppendElement(magicinterp, cp);
+		Tcl_AppendElement(magicinterp, cn);
+		if (!do_reverse)
+		    Tcl_AppendElement(magicinterp, cp);
 #else
-		TxPrintf("%s\n", cp);
+		TxPrintf("%s = \"%s\"\n", cn, cp);
 #endif
 	    }
 	    else
 	    {
 	        if (cMacro->interactive)
-		    TxPrintf("Interactive macro '%s' contains \"%s\"\n",
-			     cn, cp);
+		    TxPrintf("Interactive macro '%s' %s \"%s\"\n",
+			     cn, (do_help) ? "" : "contains", cp);
 		else
-		    TxPrintf("Macro '%s' contains \"%s\"\n", cn, cp);
+		    TxPrintf("Macro '%s' %s \"%s\"\n",
+			     cn, (do_help) ? "" : "contains", cp);
 	    }
 	    freeMagic(cn);
 	    any = TRUE;
@@ -1093,7 +1196,6 @@ windDoMacro(w, cmd, interactive)
     }
     else if (cmd->tx_argc == (argstart + 1))
     {
-	int verbose;
 	ct = MacroKey(cmd->tx_argv[argstart], &verbose);
 	if (ct == 0)
 	{
@@ -1101,7 +1203,10 @@ windDoMacro(w, cmd, interactive)
 		TxError("Unrecognized macro name %s\n", cmd->tx_argv[argstart]);
 	    return;
 	}
-	cp = MacroRetrieve(wc, ct, &iReturn);
+	if (do_help)
+	    cp = MacroRetrieveHelp(wc, ct);
+	else
+	    cp = MacroRetrieve(wc, ct, &iReturn);
 	if (cp != NULL)
 	{
 	    cn = MacroName(ct);
@@ -1143,10 +1248,32 @@ windDoMacro(w, cmd, interactive)
 	}
 	argstart++;
 	if (cmd->tx_argv[argstart][0] == '\0') MacroDelete(wc, ct);
-	else if (interactive) MacroDefine(wc, ct, cmd->tx_argv[argstart], TRUE);
-	else MacroDefine(wc, ct, cmd->tx_argv[argstart], FALSE);
+	else if (do_help)
+	    MacroDefineHelp(wc, ct, cmd->tx_argv[argstart]);
+	else if (interactive)
+	    MacroDefine(wc, ct, cmd->tx_argv[argstart], NULL, TRUE);
+	else
+	    MacroDefine(wc, ct, cmd->tx_argv[argstart], NULL, FALSE);
+	return;
+    }
+    else if (cmd->tx_argc == (argstart + 3))
+    {
+	int verbose;
+	ct = MacroKey(cmd->tx_argv[argstart], &verbose);
+	if (ct == 0)
+	{
+	    if (verbose)
+		TxError("Unrecognized macro name %s\n", cmd->tx_argv[argstart]);
+	    return;
+	}
+	argstart++;
+	if (cmd->tx_argv[argstart][0] == '\0') MacroDelete(wc, ct);
+	else if (interactive) MacroDefine(wc, ct, cmd->tx_argv[argstart],
+		cmd->tx_argv[argstart + 1], TRUE);
+	else MacroDefine(wc, ct, cmd->tx_argv[argstart],
+		cmd->tx_argv[argstart + 1], FALSE);
 	return;
     }
 
-    TxError("Usage: %s [macro_name [string]]\n", cmd->tx_argv[0]);
+    TxError("Usage: %s [macro_name [string] [help_text]]\n", cmd->tx_argv[0]);
 }
